# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import subprocess
import threading
import time
from queue import Empty, Queue


def run_and_wait(
    args=None,
    cwd=None,
    env=None,
    text=True,
    timeout=None,
    timeout_handler=None,
    output_timeout=None,
    output_timeout_handler=None,
    output_line_handler=None,
):
    """
    Run a process and wait for it to complete, with optional support for
    line-by-line output handling and timeouts.

    On timeout or output timeout, the callback should kill the process;
    many clients use  mozcrash.kill_and_get_minidump() in the timeout
    callback.

    run_and_wait is not intended to be a generic replacement for subprocess.
    Clients requiring different options or behavior should use subprocess
    directly.

    :param args: command to run. May be a string or a list.
    :param cwd: working directory for command.
    :param env: environment to use for the process (defaults to os.environ).
    :param text: open streams in text mode if True; else use binary mode.
    :param timeout: seconds to wait for process to complete before calling timeout_handler
    :param timeout_handler: function to be called if timeout reached
    :param output_timeout: seconds to wait for process to generate output
    :param output_timeout_handler: function to be called if output_timeout is reached
    :param output_line_handler: function to be called for every line of process output
    """
    is_win = os.name == "nt"

    if env is None:
        env = os.environ.copy()

    if is_win:
        kwargs = {
            "creationflags": subprocess.CREATE_NEW_PROCESS_GROUP,
        }
    else:
        kwargs = {
            "preexec_fn": os.setsid,
        }

    start_time = time.time()
    if timeout is not None:
        timeout += start_time

    queue = Queue()

    def get_line():
        queue_timeout = None
        if timeout:
            queue_timeout = timeout - time.time()
        if output_timeout:
            if queue_timeout:
                queue_timeout = min(queue_timeout, output_timeout)
            else:
                queue_timeout = output_timeout
        return queue.get(timeout=queue_timeout)

    proc = subprocess.Popen(
        args,
        cwd=cwd,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=text,
        **kwargs
    )

    def reader(fh, queue):
        for line in iter(fh.readline, "" if text else b""):
            queue.put(line)
        # Give a chance to the reading loop to exit without a timeout.
        queue.put(b"")

    threading.Thread(
        name="reader",
        target=reader,
        args=(
            proc.stdout,
            queue,
        ),
        daemon=True,
    ).start()

    try:
        for line in iter(get_line, b""):
            if output_line_handler:
                output_line_handler(proc, line)
            else:
                print(line)
        proc.wait()
    except Empty:
        if timeout and time.time() < timeout or not timeout:
            if output_timeout_handler:
                output_timeout_handler(proc)
        elif timeout_handler:
            timeout_handler(proc)

    return proc



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from telemetry_harness.fog_ping_filters import (
    FOG_DELETION_REQUEST_PING,
    FOGDocTypePingFilter,
)
from telemetry_harness.fog_testcase import FOGTestCase

FOG_DAU_REPORTING = FOGDocTypePingFilter("usage-reporting")
CANARY_USAGE_PROFILE_ID = "beefbeef-beef-beef-beef-beeefbeefbee"


class TestDauReporting(FOGTestCase):
    """Tests for FOG usage-reporting ping and dau-id cycling."""

    def test_dau_reporting(self):
        """
        Test the "usage-reporting" ping behaviour and dau-id cycling when disabling telemetry.
        """

        ping1 = self.wait_for_ping(
            self.restart_browser,
            FOG_DAU_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertNotIn("ping_info", ping1["payload"])
        self.assertNotIn("client_info", ping1["payload"])

        metrics = ping1["payload"]["metrics"]
        self.assertNotIn("legacy.telemetry.client_id", metrics["uuid"])
        self.assertNotIn("legacy.telemetry.profile_group_id", metrics["uuid"])

        self.assertIn("usage.profile_id", metrics["uuid"])
        dau_id1 = metrics["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(dau_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, dau_id1)

        # Regular `deletion-request` ping won't have the `usage.profile_id`.
        # We just wait for it to know it happened.
        _ping2 = self.wait_for_ping(
            self.disable_telemetry,
            FOG_DELETION_REQUEST_PING,
            ping_server=self.fog_ping_server,
        )

        self.enable_telemetry()
        ping3 = self.wait_for_ping(
            self.restart_browser,
            FOG_DAU_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertIn("usage.profile_id", ping3["payload"]["metrics"]["uuid"])
        dau_id2 = ping3["payload"]["metrics"]["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(dau_id2)

        self.assertNotEqual(dau_id1, dau_id2)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, dau_id2)
